/*global JQueryEventObject*/
/* Cascader.ts

	Purpose:

	Description:

	History:
		Thu Oct 3 12:43:43 CST 2019, Created by jameschu

Copyright (C) 2019 Potix Corporation. All Rights Reserved.
*/
(() => {
interface Item {
	parentId: string;
	id: string;
	content: string;
}

/**
 * A Cascader is a dropdown list in tree structure, and it supports TreeModel.
 * It does not create child widgets for each data, so the memory usage is much lower at the server.
 * <p>Available in ZK EE.
 * <p>Default {@link #getZclass}: z-cascader.
 * @author jameschu
 * @since 9.0.0
 */
zkmax.inp.Cascader = zk.$extends(zul.Widget, {
	_open: false,
	_disabled: false,
	_redrawItemsFlag: false,

	$init() {
		this.$supers('$init', arguments);
		this._items = [];
		this._itemsIdMap = [];
		this._selectedUuids = [];
	},

	$define: {
		/**
		 * Returns the placeholder text.
		 * @return String
		 */
		/**
		 * Sets the placeholder text that is displayed when selected item is empty.
		 * @param String placeholder
		 */
		placeholder(v: string) {
			let placeholder = this.$n('placeholder');
			if (placeholder) {
				placeholder.innerHTML = v;
			}
		},
		/**
		 * Returns whether it is disabled.
		 * <p>Default: false.
		 * @return boolean
		 */
		/**
		 * Sets whether it is disabled.
		 * @param boolean disabled
		 */
		disabled(v: boolean) {
			if (this.desktop) {
				jq(this.$n()).toggleClass(this.$s('disabled'))
					.attr('disabled', v ? 'disabled' : null)
					.attr('tabindex', v ? -1 : this._tabindex || 0);
			}
		},
		/**
		 * Returns the uuid list of the selected item.
		 * @return String[]
		 */
		/**
		 * Selects the items with the given uuids.
		 * @param String[] selectedUuids
		 */
		selectedUuids(selectedUuids: string[]) {
			if (this.desktop) {
				this._redrawItemsFlag = true;
				const jqLabel = jq(this.$n('label')),
					jqPlaceholder = jq(this.$n('placeholder'));
				if (selectedUuids.length == 0) {
					this._updateIconClass(this._open ? 'r' : 'd');
					jqLabel.css('display', 'none');
					jqPlaceholder.css('display', 'inline-block');
				} else {
					this._updateIconClass('x');
					jqPlaceholder.css('display', 'none');
					jqLabel.css('display', 'inline-block');
				}
			}
		},
		/**
		 * Returns the item list.
		 * @return Object[]
		 */
		/**
		 * Set the items.
		 * @param Object[] items
		 */
		items(items: Item[]) {
			const itemsIdMap = {};
			if (items == null) items = [];
			//prepare Item id map
			items.forEach((item: Item) => {
				// check parent
				let pId = item.parentId;
				if (!pId) pId = this.$class.ROOT_ID_KEY;
				let pItemInfo = itemsIdMap[pId],
					id = item.id;
				if (!pItemInfo)
					itemsIdMap[pId] = {childrenId: [id]};
				else
					pItemInfo.childrenId.push(id);
				let itemInfo = itemsIdMap[id];
				if (itemInfo)
					itemInfo['item'] = item;
				else
					itemsIdMap[id] = {item: item, childrenId: []};
			});
			this._items = items;
			this._itemsIdMap = itemsIdMap; //reset
			if (this.desktop)
				this._redrawItemsFlag = true;
		},
		/**
		 * Returns the displayed selected items text.
		 * @return String
		 */
		/**
		 * Sets the displayed selected items text.
		 * @param String label
		 */
		label(label: string) {
			if (this.desktop) {
				this.$n('label').innerHTML = label;
				jq(this.$n('label')).css('display', label ? 'inline-block' : 'none');
			}
		}
	},

	bind_() {
		this.$supers(zkmax.inp.Cascader, 'bind_', arguments);
		jq(this.$n()).on('focusin', this.proxy(this._focusinContent)).on('focusout', this.proxy(this._focusoutContent));
		jq(this.$n('pp')).on('mousedown', `.${this.$s('item')}`, this._doItemMouseDown.bind(this)).on('click', `.${this.$s('item')}`, this._doItemClick.bind(this));
		// handle icon
		if (this._lastSelectedItem) //selection is not empty
			this._updateIconClass('x');
		else {
			jq(this.$n('label')).css('display', 'none');
			jq(this.$n('placeholder')).css('display', 'inline-block');
		}
		zWatch.listen({onFloatUp: this, onResponse: this, onSize: this});
	},
	unbind_() {
		if (this._open) {
			this.close();
		}
		zWatch.unlisten({onFloatUp: this, onResponse: this, onSize: this});
		jq(this.$n('pp')).off('click', `.${this.$s('item')}`).off('mousedown', `.${this.$s('item')}`);
		jq(this.$n()).off('focusout', this.proxy(this._focusoutContent)).off('focusin', this.proxy(this._focusinContent));
		this.$supers(zkmax.inp.Cascader, 'unbind_', arguments);
	},

	domClass_(): string {
		let classes = this.$supers('domClass_', arguments);
		if (this._disabled)
			classes += ` ${this.$s('disabled')}`;
		if (this._open)
			classes += ` ${this.$s('open')}`;
		return classes;
	},

	_doItemClick(evt: JQueryEventObject) {
		this._doSelectItem(evt.currentTarget);
		evt.preventDefault();
		evt.stopPropagation();
	},
	doClick_(evt: zk.Event) {
		if (!this._disabled) {
			const domTarget = evt.domTarget;
			if (domTarget == this.$n('icon') && this._lastSelectedItem) {
				this._clearSelection();
			} else if (!jq.isAncestor(this.$n('pp'), domTarget))
				this.setOpen(!this._open);
			if (!evt.stopped)
				this.$supers('doClick_', arguments);
		}
	},

	_clearSelection() {
		this._selectedUuids = [];
		this.fire('onSelect', {selectedUuid: null});
		const selectedClass = this.$s('selected'),
			jqPp = jq(this.$n('pp'));
		jqPp.find(`.${selectedClass}`).removeClass(selectedClass);
		jqPp.find(`.${this.$s('cave')}`).eq(0).nextAll().remove();
		jq(this.$n('label')).css('display', 'none');
		jq(this.$n('placeholder')).css('display', 'inline-block');
		this._updateIconClass(this._open ? 'r' : 'd');
	},

	doKeyDown_(evt: zk.Event) {
		if (this._disabled) return;
		switch (evt.key) {
			case 'Escape':
				this.close();
				this._focusWidget();
				break;
			case 'Tab':
				this.close();
				break;
			case 'Enter':
				this._enterPressed();
				evt.stop();
				break;
			case 'ArrowUp':
				if (this._doOpenPopup(evt)) return;
				this._shiftSelection(evt, '-1');
				break;
			case 'ArrowDown':
				if (this._doOpenPopup(evt)) return;
				this._shiftSelection(evt, '1');
				break;
			case 'PageUp':
				this._shiftSelection(evt, '-30');
				break;
			case 'PageDown':
				this._shiftSelection(evt, '30');
				break;
			case 'Home':
				this._shiftSelection(evt, 'home');
				break;
			case 'End':
				this._shiftSelection(evt, 'end');
				break;
			case 'ArrowLeft':
				this._shiftSelection(evt, 'left');
				break;
			case 'ArrowRight':
				this._shiftSelection(evt, 'right');
				break;
		}
	},
	_doOpenPopup(evt: zk.Event): boolean {
		if (!this._open) {
			this.open();
			evt.stop();
			return true;
		}
		return false;
	},

	_enterPressed() {
		let jqCurrentActive = jq(this.$n('pp')).find(`.${this.$s('active')}:visible`);
		if (jqCurrentActive.length > 0)
			this._doSelectItem(jqCurrentActive.get(0), true);
	},

	_doSelectItem(itemElement: Element, isEnter?: boolean) {
		const activeClass = this.$s('active'),
			selectedClass = this.$s('selected'),
			jqItem = jq(itemElement),
			idsMap = this._itemsIdMap,
			itemId = itemElement.id,
			item = idsMap[itemId].item,
			parentId = item.parentId,
			jqCurrentCave = jqItem.parent(),
			childrenItems = this._getChildrenItems(itemId),
			selUuids = this._selectedUuids,
			selLength = selUuids.length,
			isSelected = selLength > 0 && itemId == selUuids[selLength - 1];

		jqItem.addClass(selectedClass).siblings(`.${selectedClass}`).removeClass(selectedClass);
		jq(this.$n('pp')).find(`.${activeClass}`).removeClass(activeClass); //remove active
		let finalizeSelection = true;
		if (!isEnter) {
			if (childrenItems.length > 0) {
				let nextCave = jqCurrentCave.next();
				if (nextCave.length == 0 || itemId != idsMap[nextCave.children().get(0).id].item.parentId) {
					//if next cave is empty or not its children
					jqCurrentCave.nextAll().remove();
					this._redrawItemsFlag = true; //dirty
					this._openChildrenItemPopup(itemId, jqCurrentCave[0]); //open next
					finalizeSelection = false;
				} else if (isSelected) //do nothing when non-leaf item is selected
					finalizeSelection = false;
			} else if (isSelected) //do nothing when leaf item is selected
				finalizeSelection = false;
		}
		if (finalizeSelection) {
			jqCurrentCave.nextAll().remove(); // do select
			this._lastSelectedItem = item;
		}
		const selectedUuids: string[] = []; //clear;
		selectedUuids.unshift(itemId);
		let pId = parentId;
		while (pId) {
			selectedUuids.unshift(pId);
			if (finalizeSelection)
				jq(`#${pId}`).addClass(selectedClass).siblings(`.${selectedClass}`).removeClass(selectedClass);
			pId = idsMap[pId].item.parentId;
		}
		this._selectedUuids = selectedUuids; //update selected path
		if (finalizeSelection) {
			this.close();
			this._focusWidget();
		}
	},

	_shiftSelection(evt: zk.Event, stepInfo: string) {
		let activeClass = this.$s('active'),
			itemClass = this.$s('item'),
			jqPp = jq(this.$n('pp')),
			jqCaves = jqPp.find(`.${this.$s('cave')}`),
			jqCave = jqCaves.last(),
			activeClsSelector = `.${activeClass}`,
			itemClsSelector = `.${itemClass}:visible`,
			jqCurrentActive = jqPp.find(activeClsSelector),
			jqCurrentActiveParent = jqCurrentActive.parent(),
			activeId = jqCurrentActive[0] ? jqCurrentActive[0].id : null;
		
		evt.stop();
		this._redrawItemsFlag = true; //dirty
		if (stepInfo == 'left' || stepInfo == 'right') { //handle right/left
			const itemsIdMap = this._itemsIdMap;
			if (stepInfo == 'right') {
				if (activeId && this._getChildrenItems(activeId).length > 0) {
					jqCurrentActive.removeClass(activeClass);
					const next = jqCurrentActiveParent.next();
					if (next.is(':visible')) {
						const nextCaveFirstChild = next.children(itemClsSelector).first(),
							nextCaveFirstChildId = nextCaveFirstChild.attr('id');
						if (nextCaveFirstChildId && activeId == itemsIdMap[nextCaveFirstChildId].item.parentId) {
							//jump to next layer
							nextCaveFirstChild.addClass(activeClass).get(0).scrollIntoView({block: 'nearest'});
							return;
						}
					}
					jqCurrentActiveParent.nextAll().remove(); //remove next all caves
					this._openChildrenItemPopup(activeId, jqCurrentActiveParent[0]);
					this._shiftSelection(evt, '1');
				}
			} else if (jqCaves.length > 1) { //close right most
				const activeId = jqCave.find(activeClsSelector).attr('id');
				jqCave.remove();
				if (activeId) // active item in closed cave
					jq('#' + itemsIdMap[activeId].item.parentId).addClass(activeClass);
			}
		} else {
			let jqNextActive = jqCurrentActive;
			if (stepInfo == 'home' || stepInfo == 'end') { // Home/End
				jqNextActive = jqCurrentActiveParent.children(itemClsSelector)[stepInfo == 'home' ? 'first' : 'last']();
			} else {
				let step = parseInt(stepInfo),
					isUp = step < 0,
					pos = Math.abs(step) - 1;
				if (jqCurrentActive.is(':visible')) {
					let siblings = jqCurrentActive[isUp ? 'prevAll' : 'nextAll'](':visible'),
						siblingsLength = siblings.length;
					if (siblingsLength != 0) {
						if (pos >= siblingsLength)
							pos = siblingsLength - 1; //in prevAll, the order is reversed
						jqNextActive = siblings.eq(pos);
					}
				} else {
					// No active visible
					let jqVisibleItem = jqCave.children(itemClsSelector);
					jqNextActive = isUp ? jqVisibleItem.last() : jqVisibleItem.first();
				}
			}
			if (jqCurrentActive.get(0) !== jqNextActive.get(0)) {
				jqCurrentActive.removeClass(activeClass);
				jqNextActive.addClass(activeClass).get(0).scrollIntoView({block: 'nearest'});
			}
		}
	},

	onFloatUp(ctl) {
		if (this._open && !zUtl.isAncestor(this, ctl.origin)) {
			this.close(true);
			this.fire('onBlur');
		}
	},

	onResponse() {
		if (this._redrawItemsFlag) {
			this._redrawItems();
			this._redrawItemsFlag = false;
		}
	},

	onSize() {
		if (this._open) {
			this._repositionPopup();
			this._fixBothScrollbar();
		}
	},

	_repositionPopup() {
		let pp = this.$n('pp');
		if (pp) {
			//check the popup space before position()
			let inp = this.$n(),
				$pp = zk(pp),
				ppHeight = $pp.dimension().height,
				ppWidth = $pp.dimension().width,
				inpDim = zk(inp).dimension(true),
				inpTop = inpDim.top,
				inpLeft = inpDim.left,
				inpHeight = inpDim.height,
				inpWidth = inpDim.width,
				screenX = jq.innerX(),
				screenY = jq.innerY(),
				screenHeight = jq.innerHeight(),
				screenWidth = jq.innerWidth(),
				hPosition = 'start',
				vPosition = 'after',
				opts;

			if (screenX + screenWidth - inpLeft - inpWidth > ppWidth) {
				hPosition = 'start';
			} else if (inpLeft - screenX > ppWidth) {
				hPosition = 'end';
			} else {
				opts = {overflow: true};
			}
			if (screenY + screenHeight - inpTop - inpHeight > ppHeight) {
				vPosition = 'after';
			} else if (inpTop - screenY > ppHeight) {
				vPosition = 'before';
			} else {
				opts = {overflow: true};
			}
			$pp.position(inp, vPosition + '_' + hPosition, opts);
		}
	},
	/**
	 * Drops down or closes the list of items.
	 * @param boolean open
	 */
	setOpen(open: boolean) {
		if (this.isRealVisible()) {
			if (open)
				this.open();
			else
				this.close();
		}
	},
	/**
	 * Returns whether this cascader is open.
	 *
	 * <p>Default: false.
	 * @return boolean
	 */
	isOpen(): boolean {
		return this._open;
	},

	/**
	 * Drops down the list of items.
	 */
	open() {
		if (this._open) return;
		// redraw items
		if (this._redrawItemsFlag) {
			this._redrawItems();
			this._redrawItemsFlag = false;
		}

		let pp = this.$n('pp'),
			node = this.$n(),
			zkpp = zk(pp),
			zknode = zk(node);
		jq(node).addClass(this.$s('open'));
		this._open = true;
		this.setFloating_(true, {node: pp});
		this.setFloatZIndex_(pp, this.setTopmost());
		pp.style.minWidth = jq.px0(node.offsetWidth);

		zkpp.makeVParent();
		this._repositionPopup();
		this._fixBothScrollbar();
		let fromBottom = zknode.offsetTopDoubleValue() > zkpp.offsetTopDoubleValue();
		zkpp.slideDown(this, fromBottom ? {anchor: 'b'} : {});
		
		// adjust icon direction
		if (!this._lastSelectedItem)
			this._updateIconClass('r');
		this.fire('onOpen', {open: true});
	},
	_fixBothScrollbar() {
		let zknode = zk(this.$n('cave'));
		// In non-WebKit browsers, the vertical scrollbar takes content space (overlay)
		if (zknode.hasHScroll() && zknode.hasVScroll()) {
			let jqpp = jq(this.$n('pp'));
			jqpp.width(jqpp.outerWidth()! + jq.scrollbarWidth());
		}
	},
	/**
	 * Closes the list of items.
	 */
	close(ignoreSelection?: boolean) {
		if (!this._open) return;
		if (!ignoreSelection) {
			let selItem = this._lastSelectedItem;
			if (selItem) {
				this._redrawItemsFlag = false;
				this.fire('onSelect', {selectedUuid: selItem.id});
				jq(this.$n('placeholder')).css('display', 'none');
			}
			this._lastSelectedItem = null;
		}
		jq(this.$n()).removeClass(this.$s('open'));
		this._open = false;
		const nPp = this.$n('pp'),
			activeClass = this.$s('active');

		zk(nPp).undoVParent();
		this.setFloating_(false);
		jq(nPp).find(`.${activeClass}`).removeClass(activeClass); //remove active
		// adjust icon direction
		this._updateIconClass(this._lastSelectedItem ? 'x' : 'd');
		this.fire('onOpen', {open: false});
	},

	_focusWidget() {
		var node = this.$n();
		if (!jq(node).is(':focus')) {
			this._focusByWidget = true;
			node.focus();
		}
	},

	_openChildrenItemPopup(id: string, currentCave: Element) {
		let nextSib = currentCave.nextElementSibling;
		if (!nextSib) {
			const newCave = jq('<ul/>');
			newCave.addClass(this.$s('cave'));
			jq(currentCave).after(newCave);
			nextSib = newCave[0];
		}
		const jqNextSib = jq(nextSib),
			out = new zk.Buffer();
		this._renderItems0(out, id, this._getChildrenItems(id));
		jqNextSib.html(out.join(''));
		jqNextSib.show();
	},

	_redrawpp(out: zk.Buffer) {
		out.push('<div id="', this.uuid, '-pp" class="', this.$s('popup'), ' ', this.$s('shadow'), '">');
		this._renderItems(out);
		out.push('</div>');
	},

	_renderItems(out: zk.Buffer) {
		const selectedUuids = this._selectedUuids.slice();
		selectedUuids.unshift(this.$class.ROOT_ID_KEY); //push root id
		selectedUuids.forEach((parentId: string) => {
			let items: Item[] = this._getChildrenItems(parentId);
			if (items.length > 0) {
				out.push('<ul class="', this.$s('cave'), '">');
				this._renderItems0(out, parentId, items);
				out.push('</ul>');
			}
		});
	},

	_renderItems0(out: zk.Buffer, parentId: string | null, items: Item[]) {
		let selectedUuids = this._selectedUuids,
			selectedClass = this.$s('selected'),
			itemClass = this.$s('item'),
			iconClass = this.$s('icon'),
			itemsIdMap = this._itemsIdMap;
		
		items.forEach((item: Item) => {
			let uuid = item.id,
				selected = selectedUuids.indexOf(uuid) > -1;
			out.push('<li id="', uuid, '" class="', itemClass, (selected ? ` ${selectedClass}` : ''), '">');
			out.push(item.content);
			if (itemsIdMap[uuid].childrenId.length > 0)
				out.push('<i id="', uuid, '-icon" class="', iconClass, ' z-icon-caret-right" ></i>');
			out.push('</li>');
		});
	},

	_redrawItems() {
		let out = new zk.Buffer();
		this._renderItems(out);
		jq(this.$n('pp')).html(out.join(''));
		if (this._open) {
			this._repositionPopup();
			this._fixBothScrollbar();
		}
	},

	_getChildrenItems(parentUuid: string): Item[] {
		const childrenItems: Item[] = [],
			itemsIdMap = this._itemsIdMap,
			item = itemsIdMap[parentUuid];
		if (item) {
			let childrenId = item.childrenId;
			for (let i = 0, l = childrenId.length; i < l; i++)
				childrenItems.push(itemsIdMap[childrenId[i]].item);
		}
		return childrenItems;
	},

	_updateIconClass(type: string) {
		jq(this.$n('icon')).toggleClass('z-icon-caret-down', type == 'd')
			.toggleClass('z-icon-caret-right', type == 'r')
			.toggleClass('z-icon-times', type == 'x');
	},
	_focusinContent() {
		jq(this.$n()).addClass(this.$s('focus'));
		this._skipBlur = false; // reset flag
		if (this._focusByWidget)
			this._focusByWidget = false;
		else
			this.fire('onFocus');
	},
	_focusoutContent() {
		jq(this.$n()).removeClass(this.$s('focus'));
		if (this._skipBlur) {
			this._skipBlur = false;
			return;
		}
		if (!this._focusByWidget)
			this.fire('onBlur');
	},
	_doItemMouseDown() {
		this._skipBlur = true; //to prevent blur fire event
	}
}, {
	ROOT_ID_KEY: '_z_cascader_root'
});
})();