/* Lineitem.ts

		Purpose:
		
		Description:
		
		History:
				Mon Oct 14 11:59:51 CST 2019, Created by leon

Copyright (C) 2019 Potix Corporation. All Rights Reserved.
*/

(() => {
/**
 * A lineitem.
 * <p>Lineitem is the child of Linelayout, it can contain up to 2 components inside.
 * <p>Default {@link #getZclass}: z-lineitem
 *
 * @since 9.0.0
 */
zkmax.layout.Lineitem = zk.$extends(zul.Widget, {
	_pointVisible: true,
	_pointStyle: '',
	_pointImageSrc: '',
	_opposite: false,
	_pointIconSclass: '',
	_frontSpace: '',
	_backSpace: '',

	$define: {
		/**
		 * Returns whether the point is visible.
		 * <p>Default: true.
		 * @return boolean
		 */
		/**
		 * Sets whether the point is visible.
		 *
		 * @param boolean pointVisible whether the point is visible.
		 */
		pointVisible() {
			if (this.desktop) {
				this._updatePointVisible();
			}
		},
		/**
		 * Returns the CSS style for the point.
		 * @return String
		 */
		/**
		 * Sets the pointStyle.
		 * @param String pointStyle the CSS style for the point.
		 */
		pointStyle() {
			if (this.desktop) {
				this._updatePointStyle();
			}
		},
		/**
		 * Returns the source URI of the point image.
		 * <p>Default: null.
		 * @return String
		 */
		/**
		 * Sets the source URI of the point image.
		 * @param String pointImageSrc the source URI of the point image.
		 */
		pointImageSrc() {
			if (this.desktop) {
				this._updatePointImageSrc();
			}
		},
		/**
		 * Returns whether the first child is displayed in the first area(left/top).
		 * <p>Default: false.
		 * @return boolean
		 */
		/**
		 * Sets whether the first child is displayed in the first area(left/top).
		 *
		 * @param boolean opposite whether the first child is displayed in the first area(left/top).
		 */
		opposite() {
			if (this.desktop) {
				this._updateOpposite();
			}
		},
		/**
		 * Returns the spacing with the previous lineitem.
		 * @return String
		 */
		/**
		 * Sets the spacing with the previous lineitem.
		 * @param String frontSpace the spacing with the previous lineitem. (such as "5px" or "5em").
		 * If null or empty (""), the default spacing is used (i.e.,
		 * controlled by CSS alone).
		 */
		frontSpace() {
			if (this.desktop) {
				this._updateFrontSpace();
			}
		},
		/**
		 * Returns the spacing with the next lineitem.
		 * @return String
		 */
		/**
		 * Sets the spacing with the next lineitem.
		 * @param String backSpace the spacing with the previous lineitem. (such as "5px" or "5em").
		 * If null or empty (""), the default spacing is used (i.e.,
		 * controlled by CSS alone).
		 */
		backSpace() {
			if (this.desktop) {
				this._updateBackSpace();
			}
		},
		width() {
			if (!this._nhflex) {
				let n = this.$n();
				if (n) {
					this._updateWidth();
				}
			}
		},
		height() {
			if (!this._nvflex) {
				let n = this.$n();
				if (n) {
					this._updateHeight();
				}
			}
		}
	},
	/**
	 * Returns the iconSclass name of this point.
	 * @return String
	 */
	getPointIconSclass(): string {
		return this._pointIconSclass;
	},
	/**
	 * Sets the iconSclass name of the point.
	 * @param String pointIconSclass the iconSclass name.
	 */
	setPointIconSclass(v: string) {
		if (v != this._pointIconSclass) {
			if (this.desktop) {
				this._updatePointIconSclass(v);
			}
			this._pointIconSclass = v;
		}
	},
	setVflex_(v) {
		this._updateVFlex();
	},
	setHflex_(v) {
		this._updateHFlex_();
	},
	bind_() {
		this.$supers(zkmax.layout.Lineitem, 'bind_', arguments);
		this._moveSubDOMOut();
		this._updatePointView();
		this._updateFrontSpace();
		this._updateBackSpace();
		this._updateVFlex();
		this._updateHFlex_();
		this._updateWidth();
		this._updateHeight();
	},
	unbind_() {
		if (this.desktop) {
			jq(this.$n('first')).zk.undoVParent();
			jq(this.$n('last')).zk.undoVParent();
		}
		this.$supers(zkmax.layout.Lineitem, 'unbind_', arguments);
	},
	updateOrient_() {
		this._updateFrontSpace();
		this._updateBackSpace();
		this._updateVFlex();
		this._updateHFlex_();
		this.updateScale_();
	},
	updateScale_() {
		this._updateWidth();
		this._updateHeight();
	},
	_updatePointView() {
		this._updatePointVisible();
		this._updatePointImageSrc();
		this._updatePointIconSclass(this._pointIconSclass);
		this._updatePointStyle();
	},
	_updatePointVisible() {
		const pointNode = this.$n('point');

		if (pointNode)
			jq(pointNode).toggleClass(this.$s('point-hidden'), !this._pointVisible);
	},
	_updatePointImageSrc() {
		const pointInnerNode = this.$n('point-inner');

		if (pointInnerNode)
			jq(pointInnerNode).css('background-image', 'url(' + this._pointImageSrc + ')');
	},
	_updatePointIconSclass(iconSclass: string) {
		const jqPointInnerNode = jq(this.$n('point-inner'));
		if (jqPointInnerNode.length) {
			jqPointInnerNode.removeClass(this._pointIconSclass);
			jqPointInnerNode.addClass(iconSclass);
		}
	},
	_updatePointStyle() {
		const jqPointNode = jq(this.$n('point'));

		if (jqPointNode.length) {
			zk(jqPointNode).clearStyles();
			jqPointNode.css(jq.parseStyle(this._pointStyle));
		}
	},
	_updateOpposite() {
		const firstArea = jq(this.$n('first')),
			lastArea = jq(this.$n('last')),
			lastChild = lastArea.children(); // cache children for switch area

		firstArea.children().appendTo(lastArea);
		lastChild.appendTo(firstArea);
	},
	_updateFrontSpace() {
		let frontSpace = this._frontSpace || 0,
			isParentVertical = this._isParentVertical(),
			topValue = isParentVertical ? frontSpace : 0,
			leftValue = isParentVertical ? 0 : frontSpace;

		this._cssToAllParts({'margin-top': topValue, 'margin-left': leftValue});
	},
	_updateBackSpace() {
		let backSpace = this._backSpace || 0,
			isParentVertical = this._isParentVertical(),
			bottomValue = isParentVertical ? backSpace : 0,
			rightValue = isParentVertical ? 0 : backSpace;

		this._cssToAllParts({'margin-bottom': bottomValue, 'margin-right': rightValue});
	},
	_updateWidth() {
		let width = this._width,
			flexgrow = zk.ie ? '-ms-flex-positive' : 'flex-grow';
		if (typeof width === 'undefined') return;

		if (this._isParentVertical()) {
			let finalWidth: Array<string>;
			if (!width) {
				finalWidth = ['', ''];
			} else {
				// check if 'auto' 'inherit' ... or xx% no need to calcItemSize
				finalWidth = (isNaN(parseInt(width)) || width.indexOf('%') != -1) ? [width, width] : this._calcItemSize(width);
			}
			this.$n('first').style.width = finalWidth[0];
			this.$n('last').style.width = finalWidth[1];
		} else {
			this._cssToAllParts({[flexgrow]: (width ? '0' : '1'), 'flex-basis': (width || '0')});
		}
	},
	_updateHeight() {
		let height = this._height,
			flexgrow = zk.ie ? '-ms-flex-positive' : 'flex-grow';
		if (typeof height === 'undefined') return;

		if (!this._isParentVertical()) {
			let finalHeight: Array<string>;
			if (!height) {
				finalHeight = ['', ''];
			} else {
				// check if 'auto' 'inherit' ... or xx% no need to calcItemSize
				finalHeight = (isNaN(parseInt(height)) || height.indexOf('%') != -1) ? [height, height] : this._calcItemSize(height);
			}
			this.$n('first').style.height = finalHeight[0];
			this.$n('last').style.height = finalHeight[1];
		} else {
			this._cssToAllParts({[flexgrow]: (height ? '0' : '1'), 'flex-basis': (height || '0')});
		}
	},
	_calcItemSize(size: string): string[] { // calc size for first/last parts
		const parent = this.parent;
		if (parent) {
			const pFirstScale = parent.getFirstScale(),
				pLastScale = parent.getLastScale(),
				unit = size.replace(/[0-9]/g, ''),
				eachScaleSize = parseInt(size) / (pFirstScale + pLastScale);

			return [(eachScaleSize * pFirstScale) + unit, (eachScaleSize * pLastScale) + unit];
		}
		return ['', ''];
	},
	_updateVFlex() {
		let vflex = this._vflex,
			flexgrow = zk.ie ? '-ms-flex-positive' : 'flex-grow';
		if (typeof vflex === 'undefined') return;
		let vflexIsNum = zk.parseInt(vflex) != 0;

		if (!this._isParentVertical()) {
			this._toggleStretch(vflexIsNum);
		} else {
			this._cssToAllParts({[flexgrow]: (vflexIsNum ? vflex : '1')});
		}
	},
	_updateHFlex_() {
		let hflex = this._hflex,
			flexgrow = zk.ie ? '-ms-flex-positive' : 'flex-grow';
		if (typeof hflex === 'undefined') return;
		let hflexIsNum = zk.parseInt(hflex) != 0;

		if (this._isParentVertical()) {
			this._toggleStretch(hflexIsNum);
		} else {
			this._cssToAllParts({[flexgrow]: (hflexIsNum ? hflex : '1')});
		}
	},
	_isParentVertical(): boolean {
		const parent = this.parent;
		return parent && parent.getOrient() == 'vertical';
	},
	_toggleStretch(stretch: boolean) {
		const stretchClass = this.$s('stretch');

		jq(this.$n('first')).toggleClass(stretchClass, stretch);
		jq(this.$n('last')).toggleClass(stretchClass, stretch);
	},
	_moveSubDOMOut() { // Lineitem need to move first/last part out
		const parent = this.parent,
			previousLineitem = this.previousSibling,
			firstArea = jq(this.$n('first')),
			lastArea = jq(this.$n('last'));

		if (this.desktop && parent) {
			// bind vParent before move
			firstArea.zk.makeVParent();
			lastArea.zk.makeVParent();
			if (previousLineitem) { // move to the correct place
				firstArea.insertAfter(previousLineitem.$n('first'));
				lastArea.insertAfter(previousLineitem.$n('last'));
			} else {
				firstArea.prependTo(parent.$n('first'));
				lastArea.prependTo(parent.$n('last'));
			}
		}
	},
	encloseChildHTML_(isFirstPart: boolean, out: string[]) {
		const firstChild = this.firstChild,
			lastChild = firstChild ? firstChild.nextSibling : null,
			targetChild = (this._opposite == isFirstPart) ? firstChild : lastChild;

		out.push('<div id="', this.uuid, isFirstPart ? '-first' : '-last', '" class="', this.$s(), '">');
		if (targetChild)
			targetChild.redraw(out);
		out.push('</div>');
	},
	insertChildHTML_(child, before, desktop) {
		const previousLineitem = child.previousSibling,
			opposite = this._opposite,
			firstArea = jq(this.$n('first')),
			lastArea = jq(this.$n('last'));

		if (before)
			jq(before).before(child.redrawHTML_());
		else if (opposite) {
			(previousLineitem ? lastArea : firstArea).append(child.redrawHTML_());
		} else {
			(previousLineitem ? firstArea : lastArea).append(child.redrawHTML_());
		}
		child.bind(desktop);
	},
	_cssToAllParts(css) { // do same css stuff at 3 parts
		jq(this.$n()).css(css);
		jq(this.$n('first')).css(css);
		jq(this.$n('last')).css(css);
	}
});

})();