/* Coachmark.ts

		Purpose:
		
		Description:
		
		History:
				Fri Sep 20 14:48:25 CST 2019, Created by leon

Copyright (C) 2019 Potix Corporation. All Rights Reserved.

*/
(() => {
/**
 * A coachmark component.
 * <p>A coachmark is used to attract users' attention to the target component.
 * <p>The coachmarks are suitable for guiding user operations, they should be as relevant as possible to the context.
 * <p>Only support browsers that support CSS keyframes. (IE10+, Edge, Chrome, Firefox, Safari)
 *
 * @author leon
 * @since 9.0.0
 */
zkmax.nav.Coachmark = zk.$extends(zul.Widget, {
	_target: '',
	_position: 'after_center',
	_targetOriginPosition: '',
	_targetOriginZindex: '',
	_targetHighlighted: false,

	$define: {
		/**
		 * Returns the position of this coachmark.
		 * <p>Default: after_center.
		 * @return String
		 */
		/**
		 * Sets the position of this coachmark.
		 *
		 * <p> Valid values for the position attribute are:
		 *  <ul>
		 *   <li><b>before_start</b><br/> the coachmark appears above the target, aligned to the left.</li>
		 *   <li><b>before_center</b><br/> the coachmark appears above the target, aligned to the center.</li>
		 *   <li><b>before_end</b><br/> the coachmark appears above the target, aligned to the right.</li>
		 *   <li><b>after_start</b><br/> the coachmark appears below the target, aligned to the left.</li>
		 *   <li><b>after_center</b><br/> the coachmark appears below the target, aligned to the center.</li>
		 *   <li><b>after_end</b><br/> the coachmark appears below the target, aligned to the right.</li>
		 *   <li><b>start_before</b><br/> the coachmark appears to the left of the target, aligned to the top.</li>
		 *   <li><b>start_center</b><br/> the coachmark appears to the left of the target, aligned to the middle.</li>
		 *   <li><b>start_after</b><br/> the coachmark appears to the left of the target, aligned to the bottom.</li>
		 *   <li><b>end_before</b><br/> the coachmark appears to the right of the target, aligned to the top.</li>
		 *   <li><b>end_center</b><br/> the coachmark appears to the right of the target, aligned to the middle.</li>
		 *   <li><b>end_after</b><br/> the coachmark appears to the right of the target, aligned to the bottom.</li>
		 *  </ul>
		 * </p>
		 * @param String position the position of this coachmark.
		 */
		position() {
			let n = this.$n();
			if (n)
				this._syncPosition(n);
		}
	},
	/**
	 * Returns the target component of this coachmark.
	 * @return String
	 */
	getTarget(): string {
		return this._target;
	},
	/**
	 * Sets the target component id of this coachmark.
	 * @param String target the target component id.
	 */
	setTarget(v: string) {
		if (v != this._target) {
			if (this.desktop && this._targetHighlighted)
				this._restoreTarget();
			this._target = v;
		}
	},
	setVisible(visible: boolean, fromServer?: boolean) {
		this.$supers('setVisible', arguments);
		if (this.desktop) {
			this[visible ? '_open' : '_close']();
			if (!fromServer)
				this.fire('onOpen', {open: visible});
		}
	},
	bind_() {
		this.$supers(zkmax.nav.Coachmark, 'bind_', arguments);
		zWatch.listen({afterSize: this});
		if (this.isVisible()) {
			zk.afterMount(() => {
				if (zk.webkit && 'scrollRestoration' in history && history.scrollRestoration == 'auto') { // chrome scrollRestoration issue ZK-4407
					history.scrollRestoration = 'manual';
					setTimeout(() => {
						history.scrollRestoration = 'auto';
					}, this.$class.SCROLL_RESTORATION_TIMEOUT); // turn back to 'auto' after 1000ms since we don't know when browser do scrollRestoration
				}
				this._open();
			});
		}
	},
	unbind_() {
		if (this.isVisible())
			this._close();
		zWatch.unlisten({afterSize: this});
		this.$supers(zkmax.nav.Coachmark, 'unbind_', arguments);
	},
	afterSize() {
		if (this.isVisible())
			this._syncPosition(this.$n());
	},
	_open() {
		let n = this.$n();
		if (n) {
			zk(n).makeVParent();
			this._syncPosition(n);
			this.setFloating_(true, {node: n});
			let topZIndex = this.setTopmost();
			n.style.zIndex = topZIndex > 0 ? topZIndex : 1;
			if (!this._mask) {
				let uuid = this.uuid;
				this._mask = new zk.eff.FullMask!({
					id: uuid + '-mask',
					anchor: null,
					zIndex: this._zIndex - 1,
					visible: this.isRealVisible()
				});
				jq(this.$n('mask')).addClass(this.$s('mask'));
			}
			this._highlightTarget();
			jq(n).addClass(this.$s('open'));
		}
	},
	_close() {
		let n = this.$n(),
			mask = this._mask;
		if (n) {
			this._restoreTarget();
			zk(n).undoVParent();
			this.setFloating_(false);
			if (mask) {
				mask.destroy();
				this.clearCache();
				this._mask = null;
			}
			jq(n).removeClass(this.$s('open'));
		}
	},
	_syncPosition(n: Element) {
		if (!this._visible) return;
		let targetNode = this._getTargetDom(),
			pos = this._position;
		if (targetNode) {
			this._fixPadding(this._position);
			 // the coachmark size could be changed after position,
			 // do this before scrollIntoview to prevent scrollIntoView calculate with wrong size.
			zk(n).position(targetNode, pos);
			this._scrollIntoView(targetNode);
			zk.afterAnimate(() => {
				zk(n).position(targetNode, pos); // do final position after scrollIntoview
				this._fixarrow(targetNode);
			}, -1);
		}
	},
	_fixPadding(position: string) {
		let p = this.$n('pointer');
		if (!p)
			return;
		let n = this.$n(),
			pw = jq.px0(10 + zk(p).borderWidth() / 2),
			ph = jq.px0(10 + zk(p).borderHeight() / 2);

		n.style.padding = '0';
		// cache arrow direction for _fixarrow() later
		switch (position) {
		case 'before_start':
		case 'before_center':
		case 'before_end':
			this._dir = 'd';
			n.style.paddingBottom = ph;
			break;
		case 'after_start':
		case 'after_center':
		case 'after_end':
			this._dir = 'u';
			n.style.paddingTop = ph;
			break;
		case 'end_before':
		case 'end_center':
		case 'end_after':
			this._dir = 'l';
			n.style.paddingLeft = pw;
			break;
		case 'start_before':
		case 'start_center':
		case 'start_after':
			this._dir = 'r';
			n.style.paddingRight = pw;
			break;
		default:
			this._dir = 'u';
		}
	},
	_fixarrow(targetNode: Element) {
		let p = this.$n('pointer');
		if (!p)
			return;

		let pzcls = this.$s('pointer'),
			n = this.$n(),
			dir = this._dir,
			dirMap = this.$class._DIRMAP,
			zkp = zk(p),
			pw = zkp.borderWidth(),
			ph = zkp.borderHeight(),
			nOffset = zk(n).revisedOffset(),
			targetOffset = zk(targetNode).revisedOffset(),
			arrXOffset = (targetNode.offsetWidth - pw) / 2,
			arrYOffset = (targetNode.offsetHeight - ph) / 2;
		if (dir) {
			// positioning
			if (dir == 'u' || dir == 'd') {
				let b = dir == 'u',
					l1 = (n.offsetWidth - pw) / 2,
					l2 = targetOffset[0] - nOffset[0] + arrXOffset;
				p.style.left = jq.px0(targetNode.offsetWidth >= n.offsetWidth ? l1 : l2);
				p.style[b ? 'top' : 'bottom'] = jq.px0(10 - ph / 2);
				p.style[b ? 'bottom' : 'top'] = '';
			} else {
				let b = dir == 'l',
					t1 = (n.offsetHeight - ph) / 2,
					t2 = targetOffset[1] - nOffset[1] + arrYOffset;
				p.style.top = jq.px0(targetNode.offsetHeight >= n.offsetHeight ? t1 : t2);
				p.style[b ? 'left' : 'right'] = jq.px0(10 - pw / 2);
				p.style[b ? 'right' : 'left'] = '';
			}

			p.className = pzcls + (dirMap[dir] ? ' ' + this.$s(dirMap[dir]) : '');
			jq(p).show();
		}
	},
	_getTargetDom(): Element | null {
		let targetWidget = this._smartFellow(this._target);
		return targetWidget ? targetWidget.$n() : null;
	},
	doClick_(evt: zk.Event) {
		let p = evt.domTarget;
		if (p == this.$n('cls') || p == this.$n('clsIcon'))
			this.setVisible(false);
		else
			this.$supers('doClick_', arguments);
	},
	_highlightTarget() {
		let targetNode = this._getTargetDom();
		if (targetNode && !this._targetHighlighted) {
			let targetStyle = targetNode.style,
				targetStylePosition = targetStyle.position,
				topZIndex = this._zIndex;
			this._targetOriginPosition = targetStylePosition;
			this._targetOriginZindex = targetStyle.zIndex;
			if (!targetStylePosition || (targetStylePosition == 'static'))
				targetNode.style.position = 'relative';
			targetNode.style.zIndex = topZIndex > 0 ? topZIndex : 1;
			jq(targetNode).on('click', this.proxy(this._onTargetClick));
			this._targetHighlighted = true;
		}
	},
	_restoreTarget() {
		let targetNode = this._getTargetDom();
		if (targetNode && this._targetHighlighted) {
			jq(targetNode).off('click', this.proxy(this._onTargetClick));
			targetNode.style.zIndex = this._targetOriginZindex;
			targetNode.style.position = this._targetOriginPosition;
			this._targetHighlighted = false;
		}
	},
	_onTargetClick(evt: zk.Event) {
		this.fire('onTargetClick', evt.mouseData());
	},
	_scrollIntoView(target: Element) {
		if (target) {
			const real = jq('#' + target.id + '-real')[0];
			if (real)
				target = real;
			let dir = this._dir,
				coachmark = this.$n(),
				targetOffset = jq(target).offset(),
				jqWindow = jq(window),
				doc = window || document.documentElement,
				markHeight = coachmark.offsetHeight,
				markWidth = coachmark.offsetWidth,

				targetTop = targetOffset.top - (dir == 'd' ? markHeight : 0),
				targetBottom = targetOffset.top + target.offsetHeight + (dir == 'u' ? markHeight : 0),
				windowScrollTop = jqWindow.scrollTop(),
				windowWidth = jqWindow.width(),
				totalOffsetTop = targetTop - windowScrollTop,

				targetLeft = targetOffset.left - (dir == 'r' ? markWidth : 0),
				targetRight = targetOffset.left + target.offsetWidth + (dir == 'l' ? markWidth : 0),
				windowScrollLeft = jqWindow.scrollLeft(),
				windowHeight = jqWindow.height(),
				totalOffsetLeft = targetLeft - windowScrollLeft;

			if (targetTop < windowScrollTop || targetBottom > (windowScrollTop + windowHeight)) {
				let t = target,
					offsetTop = totalOffsetTop;
				for (let p = t; (p = p.parentNode) && t != doc; t = p) {
					offsetTop = this._doScrollIntoView(p, t, offsetTop, true);
					if (offsetTop == 0) break;
				}
			}

			if (targetLeft < windowScrollLeft || targetRight > (windowScrollLeft + windowWidth)) {
				let t = target,
					offsetLeft = totalOffsetLeft;
				for (let p = t; (p = p.parentNode) && t != doc; t = p) {
					offsetLeft = this._doScrollIntoView(p, t, offsetLeft, false);
					if (offsetLeft == 0) break;
				}
			}
		}
	},
	_doScrollIntoView(outer, inner: Element, offset: number, scrollVertical: boolean): number {
		if (outer && inner) {
			if (outer == document) return offset;
			let newScrollPos = -1,
				offsetSwitch = scrollVertical ? 'top' : 'left',
				scrollSwitch = scrollVertical ? 'scrollTop' : 'scrollLeft',
				sizeSwitch = scrollVertical ? 'height' : 'width';
			if (outer != document.body) {
				//check if outer has vertical/horizontal scrollbar
				const borderSize = Math.round(jq.css(outer, scrollVertical ? 'borderLeftWidth' : 'borderTopWidth', true))
					+ Math.round(jq.css(outer, scrollVertical ? 'borderRightWidth' : 'borderBottomWidth', true)),
					scrollbarSize = outer[scrollVertical ? 'offsetWidth' : 'offsetHeight'] - borderSize - outer[scrollVertical ? 'clientWidth' : 'clientHeight'];
				if (scrollbarSize == 0) return offset;
				const toOuterOffset = jq(inner).offset()[offsetSwitch] - jq(outer).offset()[offsetSwitch];
				if (toOuterOffset == 0) return offset;

				newScrollPos = outer[scrollSwitch] + toOuterOffset;
				offset = offset - toOuterOffset;
			} else {
				outer = zk.edge_legacy ? 'body' : 'html';
				newScrollPos = jq(outer)[scrollSwitch]() + offset;
				const maxScrollPos = jq(document)[sizeSwitch]() - jq(window)[sizeSwitch]();
				if (newScrollPos > maxScrollPos) newScrollPos = maxScrollPos;
				offset = 0;
			}
			if (jq(outer)[0][scrollSwitch] != newScrollPos)
				jq(outer)[0][scrollSwitch] = newScrollPos;
			return offset;
		}
	}
}, {
	_DIRMAP: {
		'u': 'up',
		'd': 'down',
		'l': 'left',
		'r': 'right'
	},
	SCROLL_RESTORATION_TIMEOUT: 1000
});
})();